def load_dotenv(env: dict[str, str] | None = None):
    import dotenv, os

    dotenv.load_dotenv()

    if env is not None:
        for k, v in env.items():
            os.environ[k] = str(v)


from . import image, video, script

from vision_spell.types import *

__all__ = [
    "image",
    "video",
    "script",
    "Image",
    "Video",
    "Script",
    "load_dotenv",
]



import asyncio
from agentia import Agent
import nest_asyncio

from vision_spell.types import Script, ImageLike
from pydantic import BaseModel, Field


INSTRUCTIONS = """
You are an movie script writer. You will be given a brief description of the plot. Your task is to write a script for the movie or a video clip, including dialogue, scene descriptions, and any necessary actions. The script should be engaging, well-structured, and suitable for a feature-length film.

# Basics for writing prompts

Good prompts are descriptive and clear. To get your generated video closer to what you want, start with identifying your core idea and then refine your idea by adding keywords and modifiers.

The following elements should be included in your prompt:

Subject: The object, person, animal, or scenery that you want in your video.
Context: The background or context in which the subject is placed.
Action: What the subject is doing (for example, walking, running, or turning their head).
Style: This can be general or very specific. Consider using specific film style keywords, such as horror film, film noir, or animated styles like cartoon style render.
Camera motion: Optional: What the camera is doing, such as aerial view, eye-level, top-down shot, or low-angle shot.
Composition: Optional: How the shot is framed, such as wide shot, close-up, or extreme close-up.
Ambiance: Optional: How the color and light contribute to the scene, such as blue tones, night, or warm tones.
Audio: The background music, sound effects, or dialogue. Dialog should be written in the script format, with character names followed by their dialogue in quotes.

# More tips

Use descriptive language: Use adjectives and adverbs to paint a clear picture for Veo.

Provide context: If necessary, include background information to help your model understand what you want.

Reference specific artistic styles: If you have a particular aesthetic in mind, reference specific artistic styles or art movements.

Utilize prompt engineering tools: Consider exploring prompt engineering tools or resources to help you refine your prompts and achieve optimal results. For more information, see Introduction to prompting.

Enhance the facial details in your personal and group images: Specify facial details as a focus of the photo like using the word portrait in the prompt.

Clearly specify if you want audio. We recommend that you use separate sentences in your prompt to describe the audio.
"""

MODEL = "openai/gpt-4.1"


async def __script(
    description: str,
    clips: int = 1,
    dump=True,
    images: ImageLike | list[ImageLike] | None = None,
) -> Script:
    agent = Agent(model=MODEL, instructions=INSTRUCTIONS)
    duration = 8 * clips  # 8 seconds per clip
    prompt = f"""
    # NOW GENERATE A SCRIPT FOR A MOVIE BASED ON THE FOLLOWING DESCRIPTION
    # Just write the script itself, do not include any other text, no title, no description.
    # The duration of the video is {duration} seconds.

    ---

    {description}
    """
    run = agent.run(prompt, stream=True)
    content = ""
    async for stream in run:
        async for chunk in stream:
            print(chunk, end="", flush=True)
            content += chunk

    # Split into clips

    if clips > 1:
        f = Field(
            min_length=clips, max_length=clips, description="List of script clips"
        )

        class ClipsAndTitle(BaseModel):
            title: str = Field(description="The title of the whole script")
            clips: list[str] = f

        run = agent.run(
            f"""
            1. Generate a title for the overall script.
            2. Split the script into {clips} clips, each with a duration of {duration} seconds. Write a script for each clip. The clips should be connected and form a coherent story, with smooth transitions between them.
            """,
            response_format=ClipsAndTitle,
        )
        msg = await run
        cat = msg.cast(ClipsAndTitle)
        if dump:
            print("---")
            print("Title:", cat.title)
            print("---")
            for i, clip in enumerate(cat.clips):
                print(clip)
                print("---")
        clips_list = cat.clips
    else:

        class Title(BaseModel):
            title: str = Field(description="The title of the whole script")

        run = agent.run(
            "Generate a title for the overall script.", response_format=Title
        )
        msg = await run
        cat = msg.cast(Title)
        if dump:
            print("---")
            print("Title:", cat.title)
            print("---")
        clips_list = [content]

    return Script(
        title=cat.title,
        description=description,
        content=content,
        clips=clips_list,
    )


def script(
    description: str,
    clips: int = 1,
    dump=True,
    images: ImageLike | list[ImageLike] | None = None,
) -> Script:
    nest_asyncio.apply()
    return asyncio.run(__script(description, clips, dump, images))



from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import Any, Literal, Sequence, Union, TypeGuard
from IPython.display import display, Video as IPVideo, Image as IPImage
import PIL.Image
from pydantic import BaseModel, Field
import requests
from ruamel.yaml import YAML
import cv2
import tempfile
from moviepy import concatenate_videoclips, VideoFileClip
import contextlib


def _get_human_readable_size(size: int) -> str:
    s = float(size)
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if s < 1024:
            return f"{s:.2f} {unit}"
        s /= 1024
    return f"{s:.2f} PB"


@dataclass
class Video:
    bytes: bytes
    mime: str = "video/mp4"
    temporary_url: str | None = None

    def save(self, path: str | Path) -> "Video":
        """
        Save the video to a local file.
        """
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(self.bytes)
        return self

    def __repr__(self) -> str:
        s = _get_human_readable_size(len(self.bytes))
        url = f", {self.temporary_url}" if self.temporary_url else ""
        return f"Video({self.mime}, {s}{url})"

    def show(self, *, small=False) -> "Video":
        w = 240 if small else None
        h = 135 if small else None
        display(
            IPVideo(
                data=self.bytes,
                width=w,
                height=h,
                html_attributes=f"controls style='object-fit: contain; object-position: center center; background-color: black;'",
                embed=True,
                mimetype=self.mime,
            )
        )
        return self

    @staticmethod
    def load(path_or_url: "VideoLike") -> "Video":
        """
        Load a video from a file or URL.
        """
        if isinstance(path_or_url, Video):
            return path_or_url
        temporary_url = None
        if isinstance(path_or_url, Path) or not path_or_url.startswith("http"):
            path = Path(path_or_url)
            match path.suffix:
                case ".mp4":
                    mime = "video/mp4"
                case _:
                    raise ValueError(f"Unsupported video format: {path.suffix}")
            assert path.exists(), f"File not found: {path}"
            bytes = path.read_bytes()
        else:
            temporary_url = path_or_url
            response = requests.get(path_or_url)
            response.raise_for_status()
            bytes = response.content
            mime = (
                response.headers.get("Content-Type", "video/mp4")
                if not temporary_url.endswith(".mp4")
                else "video/mp4"
            )
        return Video(bytes=bytes, mime=mime, temporary_url=temporary_url)

    def __get_frame(self, pos: Literal["first", "last"]) -> "Image":
        with tempfile.NamedTemporaryFile(delete=True, suffix=".mp4") as temp_file:
            temp_file.write(self.bytes)
            vs = cv2.VideoCapture(temp_file.name)
            if pos == "first":
                vs.set(cv2.CAP_PROP_POS_FRAMES, 0)
            else:
                frames = vs.get(cv2.CAP_PROP_FRAME_COUNT)
                vs.set(cv2.CAP_PROP_POS_FRAMES, frames - 1)
            success, frame = vs.read()
            if not success:
                raise ValueError("Failed to extract last frame")
            return Image(
                bytes=cv2.imencode(".png", frame)[1].tobytes(), mime="image/png"
            )

    def get_first_frame(self) -> "Image":
        return self.__get_frame("first")

    def get_last_frame(self) -> "Image":
        return self.__get_frame("last")

    @staticmethod
    def concat(videos: Sequence[Union[str, Path, "Video"]]) -> "Video":
        with contextlib.ExitStack() as stack:
            video_clips = []
            for v in videos:
                if isinstance(v, Video):
                    f = tempfile.NamedTemporaryFile(delete=True, suffix=".mp4")
                    f.write(v.bytes)
                    f.flush()
                    stack.enter_context(f)
                    video_clips.append(VideoFileClip(f.name))
                else:
                    video_clips.append(VideoFileClip(str(v)))
            concat_clip = concatenate_videoclips(video_clips, method="compose")
            with tempfile.NamedTemporaryFile(delete=True, suffix=".mp4") as out:
                concat_clip.write_videofile(out.name)
                out.seek(0)
                return Video(bytes=out.read(), mime="video/mp4")


type VideoLike = Union[str, Path, Video]


def is_video_like(obj: Any) -> TypeGuard[VideoLike]:
    return isinstance(obj, (Video, str, Path))


@dataclass
class Image:
    bytes: bytes
    mime: str
    temporary_url: str | None = None

    def save(self, path: str | Path) -> "Image":
        """
        Save the video to a local file.
        """
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(self.bytes)
        return self

    def __repr__(self) -> str:
        s = _get_human_readable_size(len(self.bytes))
        url = f", {self.temporary_url}" if self.temporary_url else ""
        return f"Image({self.mime}, {s}{url})"

    def get_pil_image(self) -> PIL.Image.Image:
        """
        Convert the bytes to a PIL Image.
        """
        return PIL.Image.open(BytesIO(self.bytes), formats=[self.mime.split("/")[1]])

    def show(self, *, small=False) -> "Image":
        w = 240 if small else None
        h = 135 if small else None
        display(
            IPImage(
                data=self.bytes,
                width=w,
                height=h,
            )
        )
        return self

    @staticmethod
    def load(path_or_url: "ImageLike") -> "Image":
        """
        Load an image from a file or URL.
        """
        if isinstance(path_or_url, Image):
            return path_or_url
        temporary_url = None
        if isinstance(path_or_url, PIL.Image.Image):
            image = path_or_url
            bytes = BytesIO()
            image.save(bytes, format=image.format or "PNG")
            bytes.seek(0)
            mime = (
                PIL.Image.MIME.get(image.format, "image/png")
                if image.format
                else "image/png"
            )
            bytes = bytes.getvalue()
        elif isinstance(path_or_url, Path) or not path_or_url.startswith("http"):
            path = Path(path_or_url)
            assert path.exists(), f"File not found: {path}"
            bytes = path.read_bytes()
            mime = PIL.Image.open(path).get_format_mimetype() or "image/png"
        else:
            temporary_url = path_or_url
            response = requests.get(path_or_url)
            response.raise_for_status()
            bytes = response.content
            mime = response.headers.get("Content-Type", "image/png")
        return Image(bytes=bytes, mime=mime, temporary_url=temporary_url)


type ImageLike = Union[str, Path, PIL.Image.Image, Image]


def is_image_like(obj: Any) -> TypeGuard[ImageLike]:
    return isinstance(obj, (Image, PIL.Image.Image, str, Path))


class Lora(BaseModel):
    path: str
    scale: float | None = 1.0

    @staticmethod
    def __from_str(s: str) -> "Lora":
        if s.startswith("civitai:"):
            parts = s.split(":")
            assert len(parts) in (2, 3), f"Invalid civitai lora format: {s}"
            model_id = parts[1]
            url = f"https://civitai-models.0xdead.dev/{model_id}"
            scale = float(parts[2]) if len(parts) == 3 else 1.0
            return Lora(path=url, scale=scale)
        else:
            return Lora(path=s)

    @staticmethod
    def process(loras: Sequence["LoraLike"]) -> list[dict[str, Any]]:
        result = []
        for lora in loras:
            if isinstance(lora, str):
                result.append(Lora.__from_str(lora))
            else:
                result.append(lora)
        return [result.model_dump() for result in result]


type LoraLike = Union[str, Lora]


class Script(BaseModel):
    title: str
    description: str
    content: str
    clips: list[str] = Field(default_factory=list)

    def dump(self):
        print("# Title:", self.title)
        print("---")
        print("# Description\n", self.description)
        print("---")
        for i, clip in enumerate(self.clips):
            print(f"# Clip {i + 1}\n", clip)
            print("---")

    def save(self, path: str | Path) -> "Script":
        path = Path(path)
        assert path.suffix == ".yaml", "Script must be saved as a .yaml file"
        path.parent.mkdir(parents=True, exist_ok=True)

        def str_representer(dumper, data):
            if len(data.splitlines()) > 1:  # check for multiline string
                return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
            return dumper.represent_scalar("tag:yaml.org,2002:str", data)

        doc = YAML()
        doc.representer.add_representer(str, str_representer)
        doc.default_flow_style = False
        with path.open("w+") as f:
            doc.dump(self.model_dump(), f)
        return self

    @staticmethod
    def load(path: str | Path) -> "Script":
        path = Path(path)
        assert path.exists(), f"File not found: {path}"
        assert path.suffix == ".yaml", "Script must be loaded from a .yaml file"
        with path.open("r") as f:
            data = YAML().load(f)
        return Script(**data)



import os
import time
from typing import Any, Literal, Sequence
from pathlib import Path
import fal_client
from google import genai
from vision_spell.types import Video, ImageLike, VideoLike, Lora, LoraLike
from google.genai import types
from .image import __process_image_google, __process_image_falai


def load(path: str | Path) -> Video:
    """Load a video from a file path or URL."""
    return Video.load(path)


def __fal_ai_video_model(model: str, prompt: str, **kwargs: Any) -> Video:
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    handler = fal_client.subscribe(
        model,
        arguments={
            "prompt": prompt,
            "enable_safety_checker": False,
            "enable_output_safety_checker": False,
            **kwargs,
        },
    )
    url = handler["video"]["url"]
    return Video.load(url)


def hunyuan(
    prompt: str,
    *,
    image: ImageLike | None = None,
    seed: int | None = None,
    pro_mode: bool = False,
    aspect_ratio: Literal["16:9", "9:16"] | None = None,
    resolution: Literal["480p", "580p", "720p"] | None = None,
    num_frames: Literal[129, 85] | None = None,
    i2v_stability: bool | None = None,
    loras: Sequence[LoraLike] | None = None,
) -> Video:
    """
    wan2.2-a14b

    # Input Schema:

    https://fal.ai/models/fal-ai/wan-t2v/api#schema-input

    https://fal.ai/models/fal-ai/wan-i2v/api#schema-input
    """
    if image is None and not loras:
        model = "fal-ai/hunyuan-video"
    elif image and not loras:
        model = "fal-ai/hunyuan-video-image-to-video"
    elif image is None and loras:
        model = "fal-ai/hunyuan-video-lora"
    else:
        model = "fal-ai/hunyuan-video-img2vid-lora"
    image_url = __process_image_falai(image) if image else None
    return __fal_ai_video_model(
        model=model,
        prompt=prompt,
        image_url=image_url,
        seed=seed,
        pro_mode=pro_mode,
        aspect_ratio=aspect_ratio,
        resolution=resolution,
        num_frames=num_frames,
        i2v_stability=i2v_stability,
        loras=Lora.process(loras) if loras else None,
    )


def wan21(
    prompt: str,
    *,
    negative_prompt: str | None = None,
    image: ImageLike | None = None,
    num_frames: int | None = None,
    frames_per_second: int | None = None,
    seed: int | None = None,
    resolution: Literal["480p", "720p"] | None = None,
    steps: int | None = None,
    guide_scale: float | None = None,
    shift: float | None = None,
    enable_prompt_expansion: bool | None = None,
    acceleration: Literal["none", "regular"] | None = None,
    aspect_ratio: Literal["auto", "16:9", "9:16", "1:1"] | None = None,
    loras: Sequence[LoraLike] | None = None,
) -> Video:
    """
    wan2.2-a14b

    # Input Schema:

    https://fal.ai/models/fal-ai/wan-t2v/api#schema-input

    https://fal.ai/models/fal-ai/wan-i2v/api#schema-input
    """
    if image is None and not loras:
        model = "fal-ai/wan-t2v"
    elif image and not loras:
        model = "fal-ai/wan-i2v"
    elif image is None and loras:
        model = "fal-ai/wan-t2v-lora"
    else:
        model = "fal-ai/wan-i2v-lora"
    image_url = __process_image_falai(image) if image else None
    return __fal_ai_video_model(
        model=model,
        prompt=prompt,
        negative_prompt=negative_prompt,
        image_url=image_url,
        num_frames=num_frames,
        frames_per_second=frames_per_second,
        seed=seed,
        resolution=resolution,
        num_inference_steps=steps,
        guidance_scale=guide_scale,
        shift=shift,
        enable_prompt_expansion=enable_prompt_expansion,
        acceleration=acceleration,
        aspect_ratio=aspect_ratio,
        loras=Lora.process(loras) if loras else None,
    )


def wan22(
    prompt: str,
    *,
    negative_prompt: str | None = None,
    image: ImageLike | None = None,
    num_frames: int | None = None,
    frames_per_second: int | None = None,
    seed: int | None = None,
    resolution: Literal["480p", "580p", "720p"] | None = None,
    aspect_ratio: Literal["auto", "16:9", "9:16", "1:1"] | None = None,
    steps: int | None = None,
    enable_prompt_expansion: bool | None = None,
    acceleration: Literal["none", "regular"] | None = None,
    guidance_scale: float | None = None,
    guidance_scale_2: float | None = None,
    shift: float | None = None,
    interpolator_model: Literal["none", "film", "rife"] | None = None,
    num_interpolated_frames: Literal[0, 1, 2, 3, 4] | None = None,
    adjust_fps_for_interpolation: bool | None = None,
    video_quality: Literal["low", "medium", "high", "maximum"] | None = None,
    video_write_mode: Literal["fast", "balanced", "small"] | None = None,
    end_image: ImageLike | None = None,
    loras: Sequence[LoraLike] | None = None,
) -> Video:
    """
    wan2.2-a14b

    # Input Schema:

    https://fal.ai/models/fal-ai/wan/v2.2-a14b/text-to-video/api#schema-input

    https://fal.ai/models/fal-ai/wan/v2.2-a14b/image-to-video/api#schema-input
    """
    if end_image:
        assert image is not None, "end_image requires image to be provided"
    model = "fal-ai/wan/v2.2-a14b/" + (
        "text-to-video" if image is None else "image-to-video"
    )
    if loras:
        model += "/lora"
    image_url = __process_image_falai(image) if image else None
    end_image_url = __process_image_falai(end_image) if end_image else None
    return __fal_ai_video_model(
        model=model,
        prompt=prompt,
        negative_prompt=negative_prompt,
        image_url=image_url,
        num_frames=num_frames,
        frames_per_second=frames_per_second,
        seed=seed,
        resolution=resolution,
        aspect_ratio=aspect_ratio,
        num_inference_steps=steps,
        enable_prompt_expansion=enable_prompt_expansion,
        acceleration=acceleration,
        guidance_scale=guidance_scale,
        guidance_scale_2=guidance_scale_2,
        shift=shift,
        interpolator_model=interpolator_model,
        num_interpolated_frames=num_interpolated_frames,
        adjust_fps_for_interpolation=adjust_fps_for_interpolation,
        video_quality=video_quality,
        video_write_mode=video_write_mode,
        end_image_url=end_image_url,
        loras=Lora.process(loras) if loras else None,
    )


def wan25(
    prompt: str,
    *,
    negative_prompt: str | None = None,
    image: ImageLike | None = None,
    aspect_ratio: Literal["16:9", "9:16", "1:1"] | None = None,
    resolution: Literal["480p", "720p", "1080p"] | None = None,
    duration: Literal[5, 10] | None = None,
    enable_prompt_expansion: bool | None = None,
    seed: int | None = None,
) -> Video:
    """
    wan2.2-a14b

    # Input Schema:

    https://fal.ai/models/fal-ai/wan-25-preview/text-to-video/api#schema-input

    https://fal.ai/models/fal-ai/wan-25-preview/image-to-video/api#schema-input
    """
    model = "fal-ai/wan-25-preview/" + (
        "text-to-video" if image is None else "image-to-video"
    )
    if image is not None:
        assert not aspect_ratio, "aspect_ratio is not supported with image input"
    image_url = __process_image_falai(image) if image else None
    return __fal_ai_video_model(
        model=model,
        prompt=prompt,
        negative_prompt=negative_prompt,
        image_url=image_url,
        aspect_ratio=aspect_ratio,
        resolution=resolution,
        duration=duration,
        enable_prompt_expansion=enable_prompt_expansion,
        seed=seed,
    )


def sora2(
    prompt: str,
    *,
    image: ImageLike | None = None,
    resolution: Literal["720p"] | None = None,
    aspect_ratio: Literal["16:9", "9:16"] | None = None,
    duration: Literal[4, 8, 12] | None = None,
) -> Video:
    """
    sora-2

    # Input Schema:

    https://fal.ai/models/fal-ai/sora-2/text-to-video/api#schema-input

    https://fal.ai/models/fal-ai/sora-2/image-to-video/api#schema-input
    """
    model = (
        "fal-ai/sora-2/text-to-video"
        if image is None
        else "fal-ai/sora-2/image-to-video"
    )
    return __fal_ai_video_model(
        model=model,
        prompt=prompt,
        image_url=__process_image_falai(image) if image else None,
        resolution=resolution,
        aspect_ratio=aspect_ratio,
        duration=duration,
    )


def sora2_pro(
    prompt: str,
    *,
    image: ImageLike | None = None,
    resolution: Literal["720p", "1080p"] | None = None,
    aspect_ratio: Literal["16:9", "9:16"] | None = None,
    duration: Literal[4, 8, 12] | None = None,
) -> Video:
    """
    sora-2/pro

    # Input Schema:

    https://fal.ai/models/fal-ai/sora-2/text-to-video/pro/api#schema-input

    https://fal.ai/models/fal-ai/sora-2/image-to-video/pro/api#schema-input
    """
    model = (
        "fal-ai/sora-2/text-to-video/pro"
        if image is None
        else "fal-ai/sora-2/image-to-video/pro"
    )
    return __fal_ai_video_model(
        model=model,
        prompt=prompt,
        image_url=__process_image_falai(image) if image else None,
        resolution=resolution,
        aspect_ratio=aspect_ratio,
        duration=duration,
    )


def __veo3(
    prompt: str,
    *,
    negative_prompt: str | None = None,
    image: ImageLike | None = None,
    last_frame: ImageLike | None = None,
    video: VideoLike | None = None,
    fast: bool = False,
) -> Video:
    """
    Generate an image based on the provided prompt using Google's VEO3 model.
    """
    client = genai.Client(api_key=os.getenv("GEMINI_API_KEY"))

    _image = __process_image_google(image) if image else None
    _last_frame = __process_image_google(last_frame) if last_frame else None

    video = Video.load(video) if isinstance(video, (str, Path)) else video
    video_data = (
        types.Video(video_bytes=video.bytes, mime_type=video.mime)
        if isinstance(video, Video)
        else None
    )

    operation = client.models.generate_videos(
        model=(
            "veo-3.0-generate-preview" if not fast else "veo-3.0-fast-generate-preview"
        ),
        prompt=prompt,
        image=_image,
        video=video_data,
        config=types.GenerateVideosConfig(
            aspect_ratio="16:9",  # "16:9" or "9:16",
            negative_prompt=negative_prompt,
            last_frame=_last_frame,
        ),
    )

    while not operation.done:
        time.sleep(20)
        operation = client.operations.get(operation)

    if not operation.response or not operation.response.generated_videos:
        print("RESPONSE:", operation.response)
        print("ERROR:", operation.error)
        raise RuntimeError("Video generation failed")

    v = operation.response.generated_videos[0].video
    assert v, "No video generated"
    client.files.download(file=v)
    assert v.video_bytes, "No video generated"
    return Video(bytes=v.video_bytes)


def veo3(
    prompt: str,
    *,
    negative_prompt: str | None = None,
    image: ImageLike | None = None,
    last_frame: ImageLike | None = None,
    video: VideoLike | None = None,
) -> Video:
    """
    Generate an image based on the provided prompt using Google's VEO3 model.
    """
    return __veo3(
        prompt,
        negative_prompt=negative_prompt,
        image=image,
        last_frame=last_frame,
        video=video,
        fast=False,
    )


def veo3_fast(
    prompt: str,
    *,
    negative_prompt: str | None = None,
    image: ImageLike | None = None,
    last_frame: ImageLike | None = None,
    video: VideoLike | None = None,
) -> Video:
    """
    Generate an image based on the provided prompt using Google's VEO3 model.
    """
    return __veo3(
        prompt,
        negative_prompt=negative_prompt,
        image=image,
        last_frame=last_frame,
        video=video,
        fast=True,
    )


__all__ = [
    "load",
    "hunyuan",
    "wan21",
    "wan22",
    "wan25",
    "veo3",
    "veo3_fast",
]



from enum import StrEnum
from pathlib import Path
from typing import Any, Sequence, TypedDict
from typing_extensions import Literal
import fal_client
import base64
from io import BytesIO
from vision_spell.types import Image, ImageLike, Lora, LoraLike
from google.genai import types as google_genai_types


def load(path: str | Path) -> Image:
    """Load an image from a file path or URL."""
    return Image.load(path)


class ImageSize(TypedDict):
    width: int
    height: int


def __process_image_falai(image: ImageLike) -> str:
    """Process the input image and return its URL."""
    img = Image.load(image)
    image_url = fal_client.upload(img.bytes, content_type=img.mime)
    return image_url


def __process_image_google(im: ImageLike) -> google_genai_types.Image | None:
    if im is None:
        return None
    if isinstance(im, Image):
        return google_genai_types.Image(
            image_bytes=im.bytes,
            mime_type=im.mime,
        )
    return google_genai_types.Image.from_file(location=str(im))


def __fal_ai_image_model(model: str, prompt: str, **kwargs: Any) -> Image:
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    handler = fal_client.submit(
        model,
        arguments={
            "prompt": prompt,
            "sync_mode": True,
            **kwargs,
        },
    )
    response = handler.get()
    images = response.get("images", [])
    assert len(images) == 1, f"Expected 1 image, got {len(images)}"
    image_url = images[0].get("url")
    if image_url.startswith("http"):
        return Image.load(image_url)
    else:
        base64_data = image_url.split(",")[1]  # Extract the base64 data part
        mime_type = image_url.split(";")[0].split(":")[1]  # Extract the MIME type
        return Image(
            bytes=BytesIO(base64.b64decode(base64_data)).getvalue(),
            mime=mime_type or "image/png",
        )


def flex_1d(
    prompt: str,
    *,
    # square_hd, square, portrait_4_3, portrait_16_9, landscape_4_3, landscape_16_9
    image_size: (
        ImageSize
        | Literal[
            "square_hd",
            "square",
            "portrait_4_3",
            "portrait_16_9",
            "landscape_4_3",
            "landscape_16_9",
        ]
        | None
    ) = None,
    steps: int | None = None,
    seed: int | None = None,
    loras: Sequence[LoraLike] | None = None,
    guidance_scale: int | None = None,
    image: ImageLike | None = None,
) -> Image:
    """
    flex-1-dev

    # Input Schema:

    https://fal.ai/models/fal-ai/flux-lora/api#schema-input
    """
    model: str
    image_url = None
    if not image:
        model = "fal-ai/flux-lora"
        image_url = None
    else:
        model = "fal-ai/flux-lora/image-to-image"
        image_url = __process_image_falai(image)
    return __fal_ai_image_model(
        model=model,
        prompt=prompt,
        image_size=image_size,
        num_inference_steps=steps,
        seed=seed,
        loras=Lora.process(loras) if loras else None,
        guidance_scale=guidance_scale,
        image_url=image_url,
        num_images=1,
        enable_safety_checker=False,
        output_format="png",
    )


def flex_kontext(
    prompt: str,
    *,
    negative_prompt: str | None = None,
    image: ImageLike | list[ImageLike] | None = None,
    steps: int | None = None,
    guidance_scale: int | None = None,
    loras: Sequence[LoraLike] | None = None,
) -> Image:
    """
    flex-kontext

    # Input Schema:

    https://fal.ai/models/fal-ai/flux-pro/kontext/max/api#schema-input

    https://fal.ai/models/fal-ai/flux-pro/kontext/max/multi/api#schema-input

    https://fal.ai/models/fal-ai/flux-pro/kontext/max/text-to-image/api#schema-input
    """
    model: str
    image_url, image_urls = None, None
    match image:
        case _ if isinstance(image, list) and not loras:
            model = "fal-ai/flux-pro/kontext/max/multi"
            image_urls = [__process_image_falai(img) for img in image]
        case _ if image is not None and not isinstance(image, list) and not loras:
            model = "fal-ai/flux-pro/kontext/max"
            image_url = __process_image_falai(image)
        case _ if not loras:
            model = "fal-ai/flux-pro/kontext/max/text-to-image"
        case _ if not image and loras:
            model = "fal-ai/flux-kontext-lora/text-to-image"
        case _ if image and loras:
            model = "fal-ai/flux-kontext-lora"
            assert not isinstance(
                image, list
            ), "Only single image is supported with loras"
            image_url = __process_image_falai(image)
        case _:
            raise ValueError("Invalid combination of image and loras")
    return __fal_ai_image_model(
        model=model,
        prompt=prompt,
        negative_prompt=negative_prompt,
        image_url=image_url,
        image_urls=image_urls,
        num_inference_steps=steps,
        num_images=1,
        safety_tolerance=6,
        enable_safety_checker=False,
        guidance_scale=guidance_scale,
        output_format="png",
        loras=Lora.process(loras) if loras else None,
    )


def flex_kera(
    prompt: str,
    *,
    image_size: (
        Literal[
            "square_hd",
            "square",
            "portrait_4_3",
            "portrait_16_9",
            "landscape_4_3",
            "landscape_16_9",
        ]
        | ImageSize
        | None
    ) = None,
    image: ImageLike | None = None,
    strength: float | None = None,
    steps: int | None = None,
    guidance_scale: float | None = None,
    acceleration: Literal["none", "regular", "high"] | None = None,
    loras: Sequence[LoraLike] | None = None,
) -> Image:
    """
    flex-kera

    # Input Schema:

    https://fal.ai/models/fal-ai/flux/krea/api#schema-input

    https://fal.ai/models/fal-ai/flux/krea/image-to-image/api#schema-input
    """
    model: str
    image_url = None
    match image:
        case _ if image is not None and not loras:
            model = "fal-ai/flux/krea/image-to-image"
            image_url = __process_image_falai(image)
        case _ if not loras:
            model = "fal-ai/flux/krea"
        case _ if not image and loras:
            model = "fal-ai/flux-krea-lora"
        case _ if image and loras:
            model = "fal-ai/flux-krea-lora/image-to-image"
            image_url = __process_image_falai(image)
        case _:
            raise ValueError("Invalid combination of image and loras")
    return __fal_ai_image_model(
        model=model,
        prompt=prompt,
        image_url=image_url,
        num_inference_steps=steps,
        num_images=1,
        image_size=image_size,
        strength=strength,
        enable_safety_checker=False,
        guidance_scale=guidance_scale,
        acceleration=acceleration,
        output_format="png",
        loras=Lora.process(loras) if loras else None,
    )


class SDModel(StrEnum):
    SD_1_5 = "stable-diffusion-v1-5/stable-diffusion-v1-5"
    SD_2 = "stabilityai/stable-diffusion-2"
    SD_2_1 = "stabilityai/stable-diffusion-2-1"
    SD_3_5_LARGE = "stabilityai/stable-diffusion-3.5-large"
    XL_BASE_1_0 = "stabilityai/stable-diffusion-xl-base-1.0"
    XLL_DREAM_SHAPER = "Lykon/dreamshaper-xl-lightning"
    XLL_REAL_VIS = "SG161222/RealVisXL_V4.0_Lightning"

    def is_lightning(self) -> bool:
        return self in {SDModel.XLL_DREAM_SHAPER, SDModel.XLL_REAL_VIS}


def stable_diffusion(
    prompt: str,
    *,
    model: str | SDModel,
    negative_prompt: str | None = None,
    image: ImageLike | None = None,
    mask: ImageLike | None = None,
    noise_strength: float | None = None,
    loras: Sequence[LoraLike] | None = None,
    seed: int | None = None,
    # square_hd, square, portrait_4_3, portrait_16_9, landscape_4_3, landscape_16_9
    image_size: (
        Literal[
            "square_hd",
            "square",
            "portrait_4_3",
            "portrait_16_9",
            "landscape_4_3",
            "landscape_16_9",
        ]
        | ImageSize
        | None
    ) = None,
    steps: int | None = None,
    guidance_scale: float | None = None,
    clip_skip: int | None = None,
    scheduler: (
        Literal[
            "DPM++ 2M",
            "DPM++ 2M Karras",
            "DPM++ 2M SDE",
            "DPM++ 2M SDE Karras",
            "Euler",
            "Euler A",
            "Euler (trailing timesteps)",
            "LCM",
            "LCM (trailing timesteps)",
            "DDIM",
            "TCD",
        ]
        | None
    ) = None,
    sdxl_lightning: bool = False,
) -> Image:
    """
    stable-diffusion

    # Input Schema:

    https://fal.ai/models/fal-ai/lora/api#schema-input

    https://fal.ai/models/fal-ai/lora/image-to-image/api#schema-input
    """
    if sdxl_lightning:
        assert image is None, "image not supported with sdxl_lightning"
        assert mask is None, "mask not supported with sdxl_lightning"
        if isinstance(model, SDModel):
            assert model.is_lightning(), "model must be a lightning model"
        falai_model = "fal-ai/lightning-models"
    elif image is None:
        assert mask is None, "mask requires image to be provided"
        assert noise_strength is None, "noise_strength requires image to be provided"
        falai_model = "fal-ai/lora"
    elif mask is None:
        falai_model = "fal-ai/lora/image-to-image"
    else:
        falai_model = "fal-ai/lora/inpaint"
    model = model.value if isinstance(model, SDModel) else model
    if isinstance(model, str) and model.startswith("civitai:"):
        parts = model.split(":")
        assert len(parts) == 2, f"Invalid civitai model format: {model}"
        model_id = parts[1]
        model = f"https://civitai-models.0xdead.dev/{model_id}"
    return __fal_ai_image_model(
        model=falai_model,
        model_name=model,
        prompt=prompt,
        negative_prompt=negative_prompt,
        prompt_weighting=True,
        loras=Lora.process(loras) if loras else None,
        seed=seed,
        image_url=__process_image_falai(image) if image else None,
        mask_url=__process_image_falai(mask) if mask else None,
        noise_strength=noise_strength,
        image_size=image_size,
        num_inference_steps=steps,
        guidance_scale=guidance_scale,
        clip_skip=clip_skip,
        scheduler=scheduler,
        num_images=1,
        image_format="png",
        enable_safety_checker=False,
    )


def nano_banana(
    prompt: str,
    *,
    image: ImageLike | list[ImageLike] | None = None,
    aspect_ratio: Literal[
        "21:9", "1:1", "4:3", "3:2", "2:3", "5:4", "4:5", "3:4", "16:9", "9:16"
    ] = "1:1",
) -> Image:
    """
    gemini-2.5-flash-image

    # Input Schema:

    https://fal.ai/models/fal-ai/gemini-25-flash-image/api#schema-input

    https://fal.ai/models/fal-ai/gemini-25-flash-image/edit/api#schema-input
    """
    model: str
    if image is None:
        model = "fal-ai/gemini-25-flash-image"
        image_urls = None
    else:
        model = "fal-ai/gemini-25-flash-image/edit"
        if isinstance(image, list):
            image_urls = [__process_image_falai(img) for img in image]
        else:
            image_urls = [__process_image_falai(image)]
    return __fal_ai_image_model(
        model=model,
        prompt=prompt,
        image_urls=image_urls,
        num_images=1,
        output_format="png",
        aspect_ratio=aspect_ratio,
    )


__all__ = [
    "load",
    "SDModel",
    "flex_1d",
    "flex_kontext",
    "flex_kera",
    "nano_banana",
    "stable_diffusion",
]
